@@ -29,14 +29,14 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
 #include "otf.h"
 #include "otferror.h"
 
-/* Return nonzero if OTF_Glyph *G should be ignored according to
-   LookupFlag FLAG.  */
-#define IGNORED_GLYPH(g, flag)			\
-  ((g)->glyph_id == 0				\
-   || ((flag) & (1 << (g)->GlyphClass))		\
-   || (((flag) & OTF_MarkAttachmentType)	\
-       && (g)->GlyphClass == OTF_GlyphClassMark	\
-       && ((flag) >> 8) != (g)->MarkAttachClass))
+/* Return nonzero (-1 if ID is zero, 1 otherwise) if OTF_Glyph *G
+   should be ignored according to LookupFlag FLAG.  */
+#define IGNORED_GLYPH(g, flag)				\
+  ((g)->glyph_id == 0 ? -1				\
+   : (((flag) & (1 << (g)->GlyphClass))			\
+      || (((flag) & OTF_MarkAttachmentType)		\
+	  && (g)->GlyphClass == OTF_GlyphClassMark	\
+	  && ((flag) >> 8) != (g)->MarkAttachClass)))
 
 #define GSTRING_DELETE(gstring, from, len)				\
   do {									\
@@ -65,6 +65,8 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
   } while (0)
 
 
+static unsigned get_class_def (OTF_ClassDef *, OTF_GlyphID);
+
 static int
 gstring_subst (OTF *otf, OTF_GlyphString *gstring, int from, int to, int flag,
 	       OTF_GlyphID *ids, int num)
@@ -79,8 +81,11 @@ gstring_subst (OTF *otf, OTF_GlyphString *gstring, int from, int to, int flag,
   for (i = non_ignored_idx = to - 1; i >= from; i--)
     {
       OTF_Glyph *g = gstring->glyphs + i;
-      if (IGNORED_GLYPH (g, flag))
+
+      if (IGNORED_GLYPH (g, flag) == 1)
 	{
+	  /* Move this glyph to the next of the current target of
+	     substitution.  */
 	  OTF_Glyph temp = *g;
 
 	  memmove (g, g + 1, sizeof (OTF_Glyph) * (non_ignored_idx - i));
@@ -99,6 +104,12 @@ gstring_subst (OTF *otf, OTF_GlyphString *gstring, int from, int to, int flag,
     {
       gstring->glyphs[from + i].c = otf->cmap->decode_table[ids[i]];
       gstring->glyphs[from + i].glyph_id = ids[i];
+      if (otf->gdef)
+	gstring->glyphs[from + i].GlyphClass
+	  = get_class_def (&otf->gdef->glyph_class_def, ids[i]);
+      else
+	gstring->glyphs[from + i].GlyphClass = 0;
+      gstring->glyphs[from + i].positioning_type = 0;
       gstring->glyphs[from + i].f.index.from = from_idx;
       gstring->glyphs[from + i].f.index.to = to_idx;
     }
@@ -151,7 +162,8 @@ get_class_def (OTF_ClassDef *class_def, OTF_GlyphID glyph_id)
 }
 
 static OTF_LangSys *
-get_langsys (OTF_ScriptList *script_list, char *script, char *language)
+get_langsys (OTF_ScriptList *script_list,
+	     const char *script, const char *language)
 {
 
   OTF_Tag script_tag = OTF_tag (script);
@@ -189,7 +201,7 @@ get_langsys (OTF_ScriptList *script_list, char *script, char *language)
 
 static int
 setup_lookup_indices (OTF_LookupList *LookupList, OTF_FeatureList *FeatureList,
-		      char *features, int *lookup_indices)
+		      const char *features, int *lookup_indices)
 {
   int i, j, n = 0;
   OTF_Feature *feature;
@@ -473,7 +485,7 @@ lookup_gsub (OTF *otf, OTF_LookupList *lookup_list, unsigned lookup_list_index,
 	      OTF_GSUB_Alternate1 *alt1 = &subtable->u.alternate1;
 	      OTF_AlternateSet *altset = alt1->AlternateSet + coverage_idx;
 
-	      gstring_subst (otf, gstring, gidx + 1, gidx + 1, flag,
+	      gstring_subst (otf, gstring, gidx, gidx + 1, flag,
 			     altset->Alternate, altset->GlyphCount);
 	      gidx += altset->GlyphCount;;
 	    }
@@ -1339,7 +1351,8 @@ OTF_drive_gdef (OTF *otf, OTF_GlyphString *gstring)
 
 static int
 OTF_drive_gsub_internal (OTF *otf, OTF_GlyphString *gstring,
-			 char *script, char *language, char *features,
+			 const char *script, const char *language,
+			 const char *features,
 			 int alternate_subst)
 {
   char *errfmt = "GSUB driving%s";
@@ -1411,14 +1424,14 @@ OTF_drive_gsub_internal (OTF *otf, OTF_GlyphString *gstring,
 
 int
 OTF_drive_gsub (OTF *otf, OTF_GlyphString *gstring,
-		char *script, char *language, char *features)
+		const char *script, const char *language, const char *features)
 {
   return OTF_drive_gsub_internal (otf, gstring, script, language, features, 0);
 }
 
 int
 OTF_drive_gpos (OTF *otf, OTF_GlyphString *gstring,
-		char *script, char *language, char *features)
+		const char *script, const char *language, const char *features)
 {
   char *errfmt = "GPOS driving%s";
   int errret = -1;
@@ -1470,8 +1483,8 @@ OTF_drive_gpos (OTF *otf, OTF_GlyphString *gstring,
 
 int
 OTF_drive_tables (OTF *otf, OTF_GlyphString *gstring,
-		  char *script, char *language,
-		  char *gsub_features, char *gpos_features)
+		  const char *script, const char *language,
+		  const char *gsub_features, const char *gpos_features)
 {
   if (OTF_drive_cmap (otf, gstring) < 0)
     return -1;
@@ -1488,7 +1501,8 @@ OTF_drive_tables (OTF *otf, OTF_GlyphString *gstring,
 
 int
 OTF_drive_gsub_alternate (OTF *otf, OTF_GlyphString *gstring,
-			  char *script, char *language, char *features)
+			  const char *script, const char *language,
+			  const char *features)
 {
   return OTF_drive_gsub_internal (otf, gstring, script, language, features, 1);
 }
